module.exports.connected = connected;
module.exports.processMessage = processMessage;
module.exports.processSender = processSender;
module.exports.setHandlers = setHandlers;
module.exports.setSettings = setSettings;

var handlers = {
	sendHandler: null,
	findServer: null,
	findUser: null,
	killUser: null,
	newServer: null,
	newUser: null,
	parseUmodes: null,
	getChannel: null,
	quitUser: null,
	removeServer: null
};

var settings = {
	ID: null,
	password: null,
	name: null,
	description: null,
	version: null,
	me: null,
	maxUsers: null
};

var connection = {
	'protoctl': {},
	'prefixes': {},
	'chmodes': [], // 0 - list modes, 1 - argument modes, 2 - argument modes (add only), 3 - standard modes
	'umodes': {}
};

var chmodeMap = {
	'k': { 'name': 'KEY', 'user': true, 'services': true },
	'L': { 'name': 'CHANNELLINK', 'user': true, 'services': true },
	'f': { 'name': 'FLOODPROT', 'user': true, 'services': true },
	'l': { 'name': 'LIMIT', 'user': true, 'services': true },
	'H': { 'name': 'HISTORY', 'user': true, 'services': true },
	'p': { 'name': 'PRIVATE', 'user': true, 'services': true },
	's': { 'name': 'SECRET', 'user': true, 'services': true },
	'm': { 'name': 'MODERATED', 'user': true, 'services': true },
	'n': { 'name': 'NOEXTERNAL', 'user': true, 'services': true },
	't': { 'name': 'TOPIC', 'user': true, 'services': true },
	'i': { 'name': 'INVITE', 'user': true, 'services': true },
	'r': { 'name': 'REGISTERED', 'user': false, 'services': true },
	'z': { 'name': 'SECUREONLY', 'user': true, 'services': true },
	'M': { 'name': 'NONREGMODERATED', 'user': true, 'services': true },
	'Q': { 'name': 'NOKICK', 'user': true, 'services': true },
	'N': { 'name': 'NONICK', 'user': true, 'services': true },
	'R': { 'name': 'NONREGINVITE', 'user': true, 'services': true },
	'T': { 'name': 'NOCTCPS', 'user': true, 'services': true },
	'O': { 'name': 'OPERS', 'user': false, 'services': true },
	'V': { 'name': 'NOINVITE', 'user': true, 'services': true },
	'K': { 'name': 'NOKNOCK', 'user': true, 'services': true },
	'D': { 'name': 'OLDDELAYJOIN', 'user': false, 'services': false },
	'd': { 'name': 'DELAYJOIN', 'user': true, 'services': true },
	'G': { 'name': 'CENSOR', 'user': true, 'services': true },
	'P': { 'name': 'PERMANENT', 'user': false, 'services': true },
	'Z': { 'name': 'SECURE', 'user': false, 'services': false },
	'S': { 'name': 'COLORFILTER', 'user': true, 'services': true },
	'C': { 'name': 'NOCTCP', 'user': true, 'services': true },
	'c': { 'name': 'NOCOLOR', 'user': true, 'services': true }
};

function parseChannelModes(modes, args){
	var plus = true;
	var argIndex = 0;
	var output = {};
	for(var i=0; i<modes.length; i++){
		var c = modes.charAt(i);
		if(c in chmodeMap){
			var name = chmodeMap[c].name;
		} else {
			var name = c;
		}
		if(c == '+'){
			plus = true;
		} else if(c == '-'){
			plus = false;
		} else if(connection.chmodes[0].indexOf(c) >= 0){
			throw 'Unexpected list mode';
		} else if(connection.chmodes[1].indexOf(c) >= 0){
			if(plus){
				output[name] = args[argIndex++];
			} else {
				output[name] = false;
				argIndex++;
			}
		} else if(connection.chmodes[2].indexOf(c) >= 0){
			if(plus){
				output[name] = args[argIndex++];
			} else {
				output[name] = false;
			}
		} else { // treat unknown chars as type 3
			output[name] = plus;
		}
	}
	return output;
}

function parseModePrefix(text){
	var modes = [];
	for(var i=0; i<text.length; i++){
		var c = text.charAt(i);
		if(c in connection.prefixes){
			modes[connection.prefixes[c]] = true;
		} else break;
	}
	var uid = text.substring(i);
	var user = handlers.findUser(uid);
	if(!user){
		handlers.killUser(uid);
		return null;
	}
	return { 'user': user, 'modes': modes };
}

function connected(){
	ircSend(null, settings.ID, 'PASS', [settings.password]);
	ircSend(null, settings.ID, 'PROTOCTL', ['NICKv2', 'VHP', 'UMODE2', 'NICKIP', 'SJOIN', 'SJOIN2', 'SJ3', 'NOQUIT', 'TKLEXT', 'MLOCK', 'SID', 'MTAGS']);
	ircSend(null, settings.ID, 'PROTOCTL', ['EAUTH=' + settings.name + ',,,' + settings.version]);
	ircSend(null, settings.ID, 'PROTOCTL', ['SID=' + settings.ID]);
	ircSend(null, settings.ID, 'SERVER', [settings.name, '1', settings.description]);
}

function processSender(sender){
	if(sender.nick){
		if(server = handlers.findServer(sender.nick)){
			sender.nick = server.name;
			sender.server = server;
		} else if(user = handlers.findUser(sender.nick)){
			sender.nick = user.name
			sender.ident = user.ident;
			sender.host = user.host;
			sender.user = user;
		} else {
			handlers.killUser(sender.nick);
			return false;
		}
	}
	return sender;
}

function processMessage(msg){
	if(!msg) return;
	if(msg.command in cmdBinds){
		cmdBinds[msg.command](msg);
	} else {
		console.log('Unhandled cmd: '+msg.command);
		console.log(msg);
	}
}

function setHandlers(newHandlers){
	handlers = newHandlers;
}

function setSettings(newSettings){
	settings = newSettings;
}

function ircSend(tags, from, cmd, args){
	handlers.send(tags, from, cmd, args);
}

var cmdBinds = {
	'PROTOCTL': function(msg){
		for(var i=0; i<msg.args.length; i++){
			var arg = msg.args[i];
			if(arg.indexOf('=') >= 0){
				var arg = arg.split('=');
				connection.protoctl[arg[0]] = arg[1];
				switch(arg[0]){
					case 'PREFIX':
						var data = arg[1].substring(1).split(')');
						var modes = data[0].split('');
						var chars = data[1].split('');
						for(var i=0; i<modes.length; i++){
							connection.prefixes[chars[i]] = modes[i];
						}
						break;
					case 'CHANMODES':
						connection.chmodes = arg[1].split(',');
						break;
					case 'USERMODES':
						connection.umodes = arg[1];
						break;						
				}
			} else {
				connection.protoctl[arg] = true;
			}
		}
	},
	
	'SERVER': function(msg){
		var expr = /^([^-]+)-([^-]+)-([^ ]+) (.*)$/;
		var match = expr.exec(msg.args[2]);
		if(match){
			handlers.newServer(msg.args[0], match[3], match[4], msg.args[1], settings.me);
		} else {
			throw 'Unknown SERVER message';
		}
	},
	
	'MD': function(msg){
		switch(msg.args[0]){
			case 'client':
				var user = handlers.findUser(msg.args[1]);
				if(!user) return;
				switch(msg.args[2]){
					case 'certfp':
						if(!msg.args[3] || msg.args[3].length == 0) break;
						user.setSecure(true);
						user.setFingerprint(msg.args[3]);
						break;
					default: break;
				}
				break;
			default: break;
		}
	},
	
	'SMOD': function(msg){ // ignore
	},
	
	'EOS': function(msg){ // ?
	},
	
	'SINFO': function(msg){
		console.log(msg);
	},

	'SID': function(msg){
		handlers.newServer(msg.args[0], msg.args[2], msg.args[3], msg.args[1], msg.sender.server);
	},
	
	'UID': function(msg){
		var nick = msg.args[0];
		var distance = msg.args[1];
		var TS = msg.args[2];
		var ident = msg.args[3];
		var host = msg.args[4];
		var uid = msg.args[5];
		var account = msg.args[6];
		var umodes = msg.args[7];
		var vhost = msg.args[8];
		var cloakedHost = msg.args[9];
		var ip = msg.args[10];
		var realname = msg.args[11];
		if(ip != '*'){
			var ipBinary = Buffer.from(ip, 'base64');
			if(ipBinary.byteLength == 4){ // IPv4
				ip = ipBinary[0].toString(10) + '.' + ipBinary[1].toString(10) + '.' + ipBinary[2].toString(10) + '.' + ipBinary[3].toString(10);
			} else if(ipBinary.byteLength == 16){ // IPv6
				ip = '';
				for(var i=0; i<8; i++){
					if(i > 0) ip += ':';
					ip += ipBinary[i*2].toString(16).padStart(2, '0') + ipBinary[i*2+1].toString(16).padStart(2, '0');
				}
			} else {
				throw 'unknown IP format';
			}
		} else {
			ip = null;
		}
		if(vhost == '*'){
			vhost = null;
		}
		if(cloakedHost == '*'){
			cloakedHost = null;
		}
		if(account == '0'){
			account = null;
		}
		handlers.newUser(nick, distance, TS, ident, host, uid, account, handlers.parseUmodes(umodes), vhost, cloakedHost, ip, realname, msg.sender.server);
	},

	'SJOIN': function(msg){
		var channel = handlers.getChannel(msg.args[1]);
		channel.setTS(msg.args[0]);
		var modeArgs = [];
		for(var i=3; i<msg.args.length - 1; i++){
			modeArgs.push(msg.args[i]);
		}
		channel.addModes(parseChannelModes(msg.args[2], modeArgs));
		var members = msg.text.split(' ');
		for(var i=0; i<members.length; i++){
			var member = members[i];
			var c = member.charAt(0);
			if(c == '&'){ // +b
				channel.addBan(member.substring(1));
			} else if(c == '"'){ // +e
				channel.addExcept(member.substring(1));
			} else if(c == '\''){ // +I
				channel.addInvex(member.substring(1));
			} else {
				var udata = parseModePrefix(member);
				if(!udata) return;
				console.log(udata.user.name + ' joined ' + msg.args[1]);
				channel.joinUser(udata.user);
				channel.setStatusModes(udata.user, udata.modes);
			}
		}
	},
	
	'SENDUMODE': function(msg){ // ignore
	},
	
	'PASS': function(msg){ // ignore
	},
	
	'SWHOIS': function(msg){
		console.log(msg);
		//TODO
	},
	
	'MODE': function(msg){
		console.log(msg);
		//TODO
	},
	
	'TKL': function(msg){
//		console.log(msg);
		//TODO
	},
	
	'METADATA': function(msg){
		console.log(msg);
		//TODO
	},
	
	'NETINFO': function(msg){
		ircSend(null, null, 'NETINFO', [settings.maxUsers.toString(10), Math.floor(new Date() / 1000).toString(10), msg.args[2], msg.args[3], "0", "0", "0", msg.args[7]]);
	},
	
	'SASL': function(msg){
		console.log(msg);
		//TODO
	},
	
	'PING': function(msg){
		msg.reply(null, 'PONG', msg.args);
	},

	'QUIT': function(msg){
		handlers.quitUser(msg.sender.user);
	},

	'PART': function(msg){
		var channel = handlers.getChannel(msg.args[0]);
		channel.removeUser(msg.sender.user);
	},
	
	'CHGHOST': function(msg){
		var user = handlers.findUser(msg.args[0]);
		if(!user) return;
		user.changeVHost(msg.args[1]);
	},
	
	'CHGIDENT': function(msg){
		var user = handlers.findUser(msg.args[0]);
		if(!user) return;
		user.changeVIdent(msg.args[1]);
	},
	
	'CHGNAME': function(msg){
		var user = handlers.findUser(msg.args[0]);
		if(!user) return;
		user.setRealname(msg.args[1]);
	},
	
	'SETNAME': function(msg){
		msg.sender.user.setRealname(msg.args[0]);
	},
	
	'SETHOST': function(msg){
		console.log(msg);
		//TODO
	},
	
	'SETIDENT': function(msg){
		console.log(msg);
		//TODO
	},
	
	'NICK': function(msg){
		msg.sender.user.changeNick(msg.args[0]);
	},
	
	'SDESC': function(msg){
		console.log(msg);
		//TODO
	},
	
	'TOPIC': function(msg){
		console.log(msg);
		//TODO
	},
/*
messagedata {
  text: '0',
  args: [ '183.89.186.94', '0' ],
  tags: [],
  command: 'REPUTATION',
  sender: {
    nick: 'test1.pirc.pl',
    ident: '',
    host: '',
    server: server {
      name: 'test1.pirc.pl',
      sid: '093',
      description: 'serwer testowy!',
      distance: '3',
      uplink: [Object],
      introduce: [Function]
    },
    user: false
  },
  time: 2020-03-14T10:48:06.207Z,
  reply: [Function],
  originalString: ':093 REPUTATION 183.89.186.94 0'
}
*/
	'REPUTATION': function(msg){
		//TODO
	},
	
	'UMODE2': function(msg){
		console.log(msg);
		//TODO
	},
	
	'KILL': function(msg){
		var user = handlers.findUser(msg.args[0]);
		if(!user) return;
		handlers.quitUser(user);
	},

	'SQUIT': function(msg){
		var server = handlers.findServer(msg.args[0]);
		if(!server) return;
		handlers.removeServer(server);
	}
};

